/*
 *      Copyright (C) 2009 Antti Ajanki <antti.ajanki@iki.fi>
 *
 *  This Program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *
 *  This Program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with RTMPDump; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 *  http://www.gnu.org/copyleft/gpl.html
 *
 */

/*
 * This file implements the functionality from the YLE Areena media
 * player (http://areena-beta.yle.fi/player/Application.swf) that is
 * necessary for downloading streams.
 *
 * A "normal" RTMP media player extract the name of the stream
 * (playpath) from the RTMP URL, and invokes the remote procedure
 * play() on it after connecting to the server. YLE Areena has two
 * features that make downloading more difficult. When the player
 * connects to the server, the server sends a challenge and the player
 * has to respond correctly, otherwise the server refuses to talk to
 * the player. Secondly, the playpath is not available on the video
 * web page. Instead, the player has to send clipID, the final
 * component of the web page URL, to the server and extract the
 * playpath from the server's response.
 *
 * The initial communication after the handshake looks like this:
 *
 * client -> server: invoke (RTMP packet 0x14): connect()
 *
 * S -> C: invoke: authenticationDetails(..., randomAuth: 12345)
 *
 * C -> S: flex message (0x11): authenticateRandomNumber((randomAuth + 447537687) % 6834253)
 *
 * S -> C: invoke: randomNumberAuthenticated()
 *
 * C -> S: flex message: requestData("e0", "session/authenticate/1")
 *
 * C -> S: flex message: requestData("e1", "clips/info/999") (= clip ID from the URL)
 *           (or requestData("e1", "streams/info/999/lang/fi") for live streams)
 *           (or requestData("e1", "channels/info/999/lang/fi") for net radios)
 *
 * S -> C: invoke: rpcResult(..., "status": "200", ..., "url": "rtmp://host/playpath")
 *
 * C -> S: flex message: requestData("e2", "clips/related/999")
 *           (or requestData("e2", "channels/lang/fi/geo/FI") for net radios)
 *
 * C -> S: flex message: requestData("e3", "clips/featured/999")
 *           (or requestData("e3", "categories/lang/fi") for net radios)
 *
 * The e0, e2, e3 steps are not strictly required to play the stream
 * (and their results can be ignored), but they are related to some
 * auxiliary features in the Areena Flash player.
 *
 * After getting the playpath in the response to requestData("e1"),
 * the player can continue downloading the stream normally.
 *
 */

#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/nanohttp.h>

#include "json-c/json.h"

#include "yle.h"
#include "parseurl.h"
#include "akamai.h"

using namespace YLE_AREENA;

const char *defaultUserAgent = "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.9.1) Gecko/20090624 Firefox/3.5";
const char *HTTPHeaders = "User-Agent: %s\r\n"
  "Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8\r\n"
  "Accept-Language: fi,en-us;q=0.5\r\n"
  "Accept-Charset: ISO-88591-1,utf-8;q=0.7,*;q=0.7\r\n"
  "Connection: close\r\n"
  "Referer: http://areena.yle.fi/\r\n";

// Retrieve the content of url using HTTP GET method. The caller must
// free the returned buffer.
char *YLE_AREENA::GetHTTP(const char *url, const char *userAgent) {
  const char *ua = defaultUserAgent;
  if (userAgent)
    ua = userAgent;
  
  char headers[1024];
  snprintf(headers, 1024, HTTPHeaders, ua);

  // Make sure that the last two bytes are \r\n
  headers[strlen(headers)-2] = '\0';
  strcat(headers, "\r\n");

  void *httpctx = xmlNanoHTTPMethodRedir(url, "GET", NULL, NULL, NULL,
                                         headers, 0);
  if (!httpctx) {
    Log(LOGERROR, "Failed to connect to %s", url);
    return NULL;
  }

  const int inbufsize = 4096;
  int size =  0;
  int nbytes;
  char *buf = NULL;
  while (true) {
    buf = (char *)realloc(buf, (size+inbufsize+1)*sizeof(char));

    nbytes = xmlNanoHTTPRead(httpctx, &buf[size], inbufsize);
    if (nbytes <= 0)
      break;
    
    size += nbytes;
  }

  int status = xmlNanoHTTPReturnCode(httpctx);
  if (status != 200) {
    Log(LOGWARNING, "Unexpected status code from HTTP request: %d", status);
  }

  xmlNanoHTTPClose(httpctx);

  return buf;
}

// Check if there are currently live broadcasts playing. If there is,
// return URL of the video page (which the caller must free),
// otherwise return NULL.
char *YLE_AREENA::FetchLiveURL(const char *userAgent) {
  Log(LOGINFO, "Searching for live broadcasts");

  // Current and future live broadcasts are listed on this page
  const char *liveListUrl = "http://areena.yle.fi/live";
  char *buf = GetHTTP(liveListUrl, userAgent);
  if (!buf)
    return NULL;

  const char *funcstart = "AreenaPlayerAPI.openStandalonePlayer(";
  char *startptr = strstr(buf, funcstart);
  if (!startptr) {
    Log(LOGERROR, "No live streams playing at the moment.");
    free(buf);
    return NULL;
  }

  startptr += strlen(funcstart);

  char *streamID = NULL;
  char *language = NULL;

  char *saveptr;
  char *token = strtok_r(startptr, ",", &saveptr);

  if (token) {
    // Skip the first parameter; it should always be 'stream'
    token = strtok_r(NULL, ",", &saveptr);
  }

  size_t skipleft;
  size_t skipright;
  size_t len;
  if (token) {
    skipleft = strspn(token, "' ");
    skipright = 1;  // the parameter ends with '

    len = strlen(&token[skipleft]) - skipright;
    streamID = (char *)malloc((len+1)*sizeof(char));
    strncpy(streamID, &token[skipleft], len);
    streamID[len] = '\0';

    token = strtok_r(NULL, ")", &saveptr);
  }

  if (token) {
    skipleft = strspn(token, "' ");
    skipright = 1; // the parameter ends with '

    len = strlen(&token[skipleft]) - skipright;
    language = (char *)malloc((len+1)*sizeof(char));
    strncpy(language, &token[skipleft], len);
    language[len] = '\0';
  }

  char *livepage = NULL;
  if (streamID && language) {
    len = 100+strlen(streamID)+strlen(language);
    livepage = (char *)malloc((len+1)*sizeof(char));
    sprintf(livepage, "http://areena.yle.fi/player/index.php?stream=%s&language=%s", streamID, language);

    Log(LOGDEBUG, "Got live URL %s", livepage);
  } else {
    Log(LOGERROR, "Active live stream found but failed to parse its parameters!");

    char *endptr = strchr(startptr, ')');
    *endptr = '\0';
    Log(LOGERROR, "Failed parameter list was: %s", startptr);

    FREE_AND_NULL(streamID);
    FREE_AND_NULL(language);
    FREE_AND_NULL(buf);
    
    return NULL;
  }
  
  FREE_AND_NULL(streamID);
  FREE_AND_NULL(language);
  FREE_AND_NULL(buf);

  return livepage;
}

// Retrieve url, and extract the content of the <div> tag whose class
// is "AreenaPlayerParams".
char *YLE_AREENA::LoadAreenaParams(const char *url, const char *userAgent) {
  char *buf = GetHTTP(url, userAgent);
  if (!buf)
    return NULL;

  const char *starttag = "<div class=\"AreenaPlayerParams\">";
  const char *startptr = strstr(buf, starttag);
  if (!startptr) {
    Log(LOGERROR, "Can't find AreenaPlayerParams on %s.", url);
    Log(LOGERROR, "Is this really a YLE Areena video page?");
    free(buf);
    return NULL;
  }

  startptr += strlen(starttag);

  size_t paramlen;
  const char *endptr = strchr(startptr, '<');
  if (endptr)
    paramlen = endptr-startptr;
  else
    paramlen = strlen(startptr);

  char *areenaparams = (char *)malloc((paramlen+1)*sizeof(char));
  strncpy(areenaparams, startptr, paramlen);
  areenaparams[paramlen] = '\0';

  free(buf);

  Log(LOGDEBUG, "Got AreenaPlayerParams: %s", areenaparams);

  return areenaparams;
}

// Parse Areena Player parameter string. All except the first are
// output variables. Their value is changed only if the initial value
// is NULL.
void YLE_AREENA::ParseAreenaParams(const char *areenaParams, 
                                   char **clipID, 
                                   char **liveStreamID, 
                                   char **radioID,
                                   char **language, 
                                   char **hostname, 
                                   int *port, 
                                   int *protocol, 
                                   char **app)
{
  assert(clipID && liveStreamID && radioID && language && hostname && port && protocol && app);

  if (!areenaParams)
    return;

  int parsedProtocol = RTMP_PROTOCOL_UNDEFINED;
  char *parsedHost = NULL;
  unsigned int parsedPort = 0;
  char *parsedPlaypath = NULL;
  char *parsedApp = NULL;
  char *saveptr;
  char *vars = strdup(areenaParams);
  char *token = strtok_r(vars, ";", &saveptr);

  while (token) {
    //Log(LOGDEBUG, "Areena Player param: %s", token);

    if (!*clipID && strncmp(token, "clip:", 5)==0) {
      *clipID = strdup(&token[5]);
    } else if (!*liveStreamID && strncmp(token, "stream:", 7)==0) {
      *liveStreamID = strdup(&token[7]);
    } else if (!*radioID && strncmp(token, "channel:", 8)==0) {
      *radioID = strdup(&token[8]);
    } else if (!*language && strncmp(token, "language:", 9)==0) {
      *language = strdup(&token[9]);
    } else if (!strncmp(token, "fmsurl:", 7)) {
      if (ParseUrl(&token[7], &parsedProtocol, &parsedHost, 
                   &parsedPort, &parsedPlaypath, &parsedApp)) {

        if ((*protocol == RTMP_PROTOCOL_UNDEFINED) || (*protocol == -1))
          *protocol = parsedProtocol;
        if (parsedHost && !*hostname)
          *hostname = strdup(parsedHost);
        if (*port == -1)
          *port = parsedPort;
        if (parsedApp && !*app)
          *app = strdup(parsedApp);

        if (parsedPlaypath && *parsedPlaypath)
          Log(LOGWARNING, "Ignoring playpath component of the fmsurl in Areena Player parameters.");

        FREE_AND_NULL(parsedHost);
        FREE_AND_NULL(parsedPlaypath);
        FREE_AND_NULL(parsedApp);
      } else {
        Log(LOGWARNING, "Failed to parse fmsurl in Areena Player parameters.");
      }
       
    }

    token = strtok_r(NULL, ";", &saveptr);
  }

  free(vars);
}

AreenaLink::AreenaLink() {
  clipID = NULL;
  liveStreamID = NULL;
  radioID = NULL;
  language = NULL;
  tvFeeRequired = false;
  tvFeeActivated = false;
  locatedInBroadcastTerritory = true;;
  isAkamaiStream = false;
}

AreenaLink::~AreenaLink() {
  Reset();
}

void AreenaLink::Reset() {
  FREE_AND_NULL(clipID);
  FREE_AND_NULL(liveStreamID);
  FREE_AND_NULL(radioID);
  FREE_AND_NULL(language);
}

YLERTMP::YLERTMP() {
  fileextension = NULL;
  m_fAudioCodecs = 3191.0;
  m_fVideoCodecs = 252.0;
}

YLERTMP::~YLERTMP() {
  if (fileextension)
    free(fileextension);
}

 void YLERTMP::SetupStream(
      int protocol, 
      const char *hostname, 
      unsigned int port, 
      const char *playpath, 
      const char *tcUrl, 
      const char *swfUrl, 
      const char *pageUrl, 
      const char *app, 
      const char *clipID,
      const char *liveStreamID,
      const char *radioID,
      const char *language,
      const char *auth,
      const char *swfSHA256Hash,
      uint32_t swfSize,
      const char *flashVer, 
      double dTime,
      bool bLiveStream,
      long int timeout)
{
  // If more than one of clip, live stream or radio IDs are
  // specified, prefer live stream to clip, and clip to radio
  if (liveStreamID && clipID) {
    Log(LOGDEBUG, "Live stream overrides clip.");
    clipID = NULL;
  }
  if (clipID && radioID) {
    Log(LOGDEBUG, "Clip overrides radio.");
    radioID = NULL;
  }
  if (liveStreamID && radioID) {
    Log(LOGDEBUG, "Live stream overrides radio.");
    radioID = NULL;
  }

  areenaStream.Reset();
  areenaStream.clipID = clipID ? strdup(clipID) : NULL;
  areenaStream.liveStreamID = liveStreamID ? strdup(liveStreamID) : NULL;
  areenaStream.radioID = radioID ? strdup(radioID) : NULL;
  areenaStream.language = language ? strdup(language) : NULL;

  // The AreenaPlayer SWF uses string "1" as the default value for
  // auth parameter if no other value is given. (An empty values seems
  // to work also.)
  const char *yleauth = auth;
  if (!auth)
    yleauth = "1";

  CRTMP::SetupStream(protocol, hostname, port, playpath, tcUrl, \
                     swfUrl, pageUrl, app, yleauth, swfSHA256Hash, \
                     swfSize, flashVer, dTime, bLiveStream, timeout);
}

bool YLERTMP::Connect() {
  if ((Link.bLiveStream && !areenaStream.liveStreamID) || 
      (!Link.bLiveStream && !areenaStream.clipID && !areenaStream.radioID))
    return false;

  Log(LOGDEBUG, "clip ID  : %s", areenaStream.clipID);
  Log(LOGDEBUG, "live ID  : %s", areenaStream.liveStreamID);
  Log(LOGDEBUG, "radio ID : %s", areenaStream.radioID);
  Log(LOGDEBUG, "language : %s", areenaStream.language);

  return CRTMP::Connect();
}

void YLERTMP::ExecuteInvokedMethod(const RTMP_LIB::AMFObject& obj) {
  std::string method = obj.GetProperty(0).GetString();
  Log(LOGDEBUG, "YLERTMP::%s, server invoking <%s>", __FUNCTION__, method.c_str());

  bool success = true;
  if ((method == "_result") && (m_methodCalls[0] == "connect"))  {
    // Override connect event handler. The base class invokes
    // CreateStream as response to connect event. We don't want to do
    // that just yet, but instead we need to first complete the
    // authentication and ask for the clip URL.
    m_methodCalls.erase(m_methodCalls.begin());

    Log(LOGDEBUG, "%s, received result for method call <connect>", __FUNCTION__);

    SendServerBW();
    //SendPing(3, 0, 300);

  } else if (method == "authenticationDetails") {
    success = ExecuteAuthenticationDetails(obj.GetProperty(3).GetObject());

  } else if (method == "randomNumberAuthenticated") {
    success = ExecuteRandomNumberAuthenticated();

  } else if (method == "rpcResult") {
    success = ExecuteRPCResult(obj);

  } else if (method == "rpcError") {
    Log(LOGERROR, "rtmp server sent error");
    success = false;
    
  } else {
    CRTMP::ExecuteInvokedMethod(obj);
  }

  if (!success) {
    Log(LOGDEBUG, "ExecuteInvokedMethod (%s) returned false!", method.c_str());
    Close();
  }
}

bool YLERTMP::ExecuteAuthenticationDetails(const RTMP_LIB::AMFObject& properties) {
  long authResult = 0;
  bool got_randomAuth = false;
  for (int i=0; i<properties.GetPropertyCount(); i++) {
    RTMP_LIB::AMFObjectProperty p = properties.GetProperty(i);
    std::string name = p.GetPropName();

    if (name == "locatedInBroadcastTerritory") {
      areenaStream.locatedInBroadcastTerritory = p.GetBoolean();
    } else if (name == "randomAuth") {
      authResult = ((long)p.GetNumber() + 447537687) % 6834253;
      got_randomAuth = true;
    } else if (name == "tvFeeActivated") {
      areenaStream.tvFeeActivated = p.GetBoolean();
    }
  }

  if (got_randomAuth) {
    RTMP_LIB::RTMPPacket packet;
    packet.m_nChannel = 0x03; // control channel
    packet.m_headerType = RTMP_PACKET_SIZE_LARGE;
    packet.m_packetType = 0x11; // FLEX MESSAGE
    packet.m_nInfoField1 = GetTime();
    
    if (!packet.AllocPacket(47))
      return false;

    char *enc = packet.m_body;
    *enc = 0x00;   // Unknown
    enc++;
    enc += EncodeString(enc, "authenticateRandomNumber");
    enc += EncodeNumber(enc, 0);
    *enc = 0x05;   // NULL
    enc++;
    enc += EncodeNumber(enc, (double)authResult);

    packet.m_nBodySize = enc-packet.m_body;
    
    assert(packet.m_nBodySize == 47);

    return SendRTMP(packet);
  }

  return false;
}

bool YLERTMP::ExecuteRandomNumberAuthenticated() {
  // It seems to be possible to skip SessionAuthenticate() and do
  // ClipsInfo() directly, but let's follow the Flash player as
  // closely as possible.

  return SessionAuthenticate();
  //return ClipsInfo();
}

bool YLERTMP::RequestData(const std::string& request, const std::string& id) {
  RTMP_LIB::RTMPPacket packet;
  packet.m_nChannel = 0x03; // control channel
  packet.m_headerType = RTMP_PACKET_SIZE_LARGE;
  packet.m_packetType = 0x11; // FLEX MESSAGE
  packet.m_nInfoField1 = GetTime();
  
  unsigned int bodylen = 31 + request.length() + id.length();
  if (!packet.AllocPacket(bodylen))
    return false;

  char *enc = packet.m_body;
  *enc = 0x00;   // Unknown
  enc++;
  enc += EncodeString(enc, "requestData");
  enc += EncodeNumber(enc, 0);
  *enc = 0x05;   // NULL
  enc++;
  enc += EncodeString(enc, id);
  enc += EncodeString(enc, request);

  packet.m_nBodySize = enc-packet.m_body;
  
  assert(packet.m_nBodySize == bodylen);
  
  Log(LOGDEBUG, "Sending requestData (%s) %s", id.c_str(), request.c_str());

  return SendRTMP(packet);
}

bool YLERTMP::SessionAuthenticate() {
  std::string authstr("session/authenticate/");
  if (Link.auth)
    authstr += Link.auth;
  else
    authstr += "1";

  return RequestData(authstr, "e0");
}

bool YLERTMP::ClipsInfo() {
  std::string cmd;
  if (areenaStream.liveStreamID) {
    cmd = "streams/info/";
    cmd += areenaStream.liveStreamID;
    cmd += "/lang/";
    if (areenaStream.language)
      cmd += areenaStream.language;
  } else if (areenaStream.clipID) {
    cmd = "clips/info/";
    cmd += areenaStream.clipID;
  } else if (areenaStream.radioID) {
    cmd = "channels/info/";
    cmd += areenaStream.radioID;
    cmd += "/lang/";
    if (areenaStream.language)
      cmd += areenaStream.language;
  } else {
    return false;
  }

  return RequestData(cmd, "e1");
}

bool YLERTMP::RelatedClips() {
  if (!areenaStream.clipID)
    return false;

  std::string cmd("clips/related/");
  cmd += areenaStream.clipID;
  return RequestData(cmd, "e2");
}

bool YLERTMP::FeaturedClips() {
  if (!areenaStream.clipID)
    return false;

  std::string cmd("clips/featured/");
  cmd += areenaStream.clipID;
  return RequestData(cmd, "e3");
}

bool YLERTMP::ChannelsList() {
  const char *lang = areenaStream.language;
  if (!lang)
    lang = "fi";

  std::string cmd = "channels/lang/";
  cmd += lang;
  cmd += "/geo/FI";

  return RequestData(cmd, "e2");
}

bool YLERTMP::CategoriesList() {
  const char *lang = areenaStream.language;
  if (!lang)
    lang = "fi";

  std::string cmd = "categories/lang/";
  cmd += lang;

  return RequestData(cmd, "e3");
}

bool YLERTMP::ExecuteRPCResult(const RTMP_LIB::AMFObject& obj) {
  bool res = false;
  int jsonstatus = -1;
  std::string callID = obj.GetProperty(3).GetString();
  std::string jsondata = obj.GetProperty(4).GetString();
  char *message = strdup(jsondata.c_str());
  struct json_object *json = json_tokener_parse(message);
  free(message);

  if (!is_error(json)) {
    json_object_object_foreach(json, outerkey, outerval) {
      if (!strcmp(outerkey, "status")) {
        json_object_object_foreach(outerval, key, val) {
          if (!strcmp(key, "code")) {
            jsonstatus = json_object_get_int(val);
          }
        }
      }
    }

    if (jsonstatus != 200) {
      Log(LOGWARNING, "Unexpected RPC status code: %d. Trying to continue anyway.", jsonstatus);
    }
  } else if (jsondata != "") {
    Log(LOGWARNING, "Failed to parse JSON string: %s", jsondata.c_str());
  }

  //Log(LOGDEBUG, "response to %s:", callID.c_str());
  //Log(LOGDEBUG, jsondata.c_str());

  if (callID == "e0") {
    // Response ignored
    res = ClipsInfo();

  } else if ((callID == "e1") && !is_error(json)) { // clips/info/12345
    char *title = NULL;
    char *seriesTitle = NULL;
    char *date = NULL;
    char *publicationDate = NULL;

    json_object_object_foreach(json, outerkey, outerval) {
      if (!strcmp(outerkey, "data")) {
        json_object_object_foreach(outerval, key, val) {

          if (!strcmp(key, "url")) {
            char *playurl = json_object_get_string(val);
            if (playurl) {
              if (Link.playpath)
                free(Link.playpath);
              
              Link.playpath = UrlToPlaypath(playurl);
              res = true;
              
              Log(LOGDEBUG, "Updated playpath: %s", Link.playpath);
            }

          } else if (!strcmp(key, "seriesTitle")) {
            char *jsontitle = json_object_get_string(val);
            if (jsontitle && *jsontitle) {
              seriesTitle = strdup(jsontitle);
              Log(LOGDEBUG, "Got seriesTitle: %s", seriesTitle);
            }

          } else if (!strcmp(key, "title")) {
            char *jsontitle = json_object_get_string(val);
            if (jsontitle && *jsontitle) {
              title = strdup(jsontitle);
              Log(LOGDEBUG, "Got title: %s", title);
            }

          } else if (!strcmp(key, "broadcastDateTime")) {
            char *jsondate = json_object_get_string(val);
            if (jsondate && (strcmp(jsondate, "0") != 0)) {
              date = strdup(jsondate);
              Log(LOGDEBUG, "Got broadcastDateTime: %s", date);
            }

          } else if (!strcmp(key, "publicationDateTime")) {
            char *jsondate = json_object_get_string(val);
            if (jsondate) {
              publicationDate = strdup(jsondate);
              Log(LOGDEBUG, "Got publicationDateTime: %s", publicationDate);
            }

          } else if (!strcmp(key, "akamai")) {
            areenaStream.isAkamaiStream = json_object_get_boolean(val);
            if (areenaStream.isAkamaiStream) {
              Log(LOGDEBUG, "This is an Akamai stream");
            }

          } else if (!strcmp(key, "tvFeeRequired")) {
            areenaStream.tvFeeRequired = json_object_get_boolean(val);
          }
        }
      }
    }

    if (res) {
      if (seriesTitle) {
        streamTitle = std::string(seriesTitle);
      } else if (title) {
        streamTitle = std::string(title);
      } else {
        streamTitle = std::string("YLEAreena-");
        if (areenaStream.liveStreamID) {
          streamTitle += "live-";
          streamTitle += areenaStream.liveStreamID;
        } else if (areenaStream.clipID) {
          streamTitle += "clip-";
          streamTitle += areenaStream.clipID;
        } else {
          streamTitle += "radio-";
          streamTitle += areenaStream.radioID;
        }
      }

      if (date && *date) {
        streamTitle += "-";
        streamTitle += date;
      } else if (publicationDate && *publicationDate) {
        streamTitle += "-";
        streamTitle += publicationDate;
      }

      if (areenaStream.isAkamaiStream) {
        LoadAkamaiXML(Link.playpath);

        // Abort this connection because the actual stream is on a
        // different server.
        res = false; 
      } else {
        if (areenaStream.clipID)
          res = RelatedClips() && FeaturedClips();
        else if (areenaStream.radioID)
          res = ChannelsList() && CategoriesList();
        else
          res = SendCreateStream(2.0);
      }
    }

    FREE_AND_NULL(title);
    FREE_AND_NULL(seriesTitle);
    FREE_AND_NULL(date);
    FREE_AND_NULL(publicationDate);

  } else if (callID == "e2") {
    // Response ignored
    res = true;

  } else if (callID == "e3") {
    // Response ignored
    res = SendCreateStream(2.0);

  }

  if (json && !is_error(json))
    json_object_put(json);

  return res;
}

char *YLERTMP::UrlToPlaypath(const char *url) {
  if (!url)
    return NULL;

  // Skip rtmp://server/app/
  const char *src = url;
  if (strncmp(src, "rtmp://", 7) == 0) {
    src += 7;
    for (int i=0; i<2; i++) {
      src = strchr(src, '/');
      if (!src)
        break;
      src++;
    }
  }
  if (!src)
    src = url;

  char *playpath = ParsePlaypath(src);

  // The stream is always Flash video regardless of the extension in
  // the URL.
  if (fileextension)
    free(fileextension);
  fileextension = strdup(".flv");

  Log(LOGDEBUG, "%s: %s -> %s", __FUNCTION__, url, playpath);

  return playpath;
}

std::string YLERTMP::GetSuggestedFileName() {
  std::string name;

  if (fileextension)
    name = streamTitle + fileextension;
  else
    name = streamTitle;

  return SaneFileName(name);
}

std::string YLERTMP::SaneFileName(const std::string& filename) {
  std::string newname = filename;

  // Remove dots from the beginning.
  while (newname[0] == '.')
    newname.erase(0, 1);

  // Windows does not accept names that end in a dot.
  while (newname[newname.length()-1] == '.')
    newname.erase(newname.length()-1, 1);

  // The following characters are illegal on some filesystems ('/' on
  // Unix filesystems and the rest on NTFS and on VFAT). Replace them
  // with an underscore.
  const char *excludechr = "\"*/:<>?\\|";

  size_t iex = newname.find_first_of(excludechr, 0);
  while (iex != newname.npos) {
    newname[iex] = '_';
    iex = newname.find_first_of(excludechr, iex+1);
  }

  return newname;
}

RTMP_LIB::CRTMP *YLERTMP::GetRedirection() {
  if (areenaStream.isAkamaiStream) {
    AKAMAI::AkamaiRTMP *akamai = new AKAMAI::AkamaiRTMP();
    akamai->SetupStream(akamaiStream.protocol,
                        akamaiStream.hostname,
                        akamaiStream.port,
                        akamaiStream.playpath,
                        akamaiStream.tcUrl,
                        akamaiStream.swfUrl,
                        akamaiStream.pageUrl,
                        akamaiStream.app,
                        akamaiStream.auth,
                        akamaiStream.SWFHash,
                        akamaiStream.SWFSize,
                        akamaiStream.flashVer,
                        akamaiStream.seekTime,
                        akamaiStream.bLiveStream,
                        akamaiStream.timeout);
    return akamai;
  } else {
    return NULL;
  }
}

bool YLERTMP::TVFeeRequiredAndMissing() {
  return areenaStream.tvFeeRequired && !areenaStream.tvFeeActivated;
}

void YLERTMP::LoadAkamaiXML(const char *url) {
  char *contenttype = NULL;
  char *redir = NULL;

  char *httpurl = strdup(url);
  if (strncmp(httpurl, "rtmp://", 7) == 0) {
    strncpy(httpurl, "http", 4); // do not copy the terminating null-character!
  }

  // Read XML from server
  void *httpctx = xmlNanoHTTPOpenRedir(httpurl, &contenttype, &redir);

  if (contenttype)
    free(contenttype);
  if (redir)
    free(redir);

  const int inbufsize = 4096;
  int size =  0;
  int nbytes;
  char *buf = NULL;
  while (true) {
    buf = (char *)realloc(buf, (size+inbufsize+1)*sizeof(char));

    nbytes = xmlNanoHTTPRead(httpctx, &buf[size], inbufsize);
    if (nbytes <= 0)
      break;
    
    size += nbytes;
  }

  int httpcode = xmlNanoHTTPReturnCode(httpctx);
  xmlNanoHTTPClose(httpctx);

  if (nbytes < 0 || !buf) {
    Log(LOGERROR, "Error reading Akamai XML (error code = %d).", httpcode);
    free(httpurl);
    if (buf)
      free(buf);
    return;
  }

  buf[size] = '\0';

  Log(LOGDEBUG, "Akamai XML:");
  Log(LOGDEBUG, "%s", buf);

  ParseAkamaiXML(buf, httpurl);

  free(buf);
  free(httpurl);
}

void YLERTMP::ParseAkamaiXML(const char *xml, const char *xmlurl) {
  if (!xml)
    return;

  xmlChar *ip = NULL;
  xmlChar *hostname = NULL;
  xmlChar *appName = NULL;
  xmlChar *streamName = NULL;
  xmlChar *authParams = NULL;
  xmlDoc *doc = xmlReadMemory(xml, strlen(xml), xmlurl, NULL,   \
                              XML_PARSE_RECOVER | \
                              XML_PARSE_NOERROR | \
                              XML_PARSE_NOWARNING | \
                              XML_PARSE_NONET);
  xmlNode *node = xmlDocGetRootElement(doc);

  node = node->children;
  while (node) {
    if (xmlStrEqual(node->name, BAD_CAST "streamName")) {
      if (streamName) xmlFree(streamName);
      streamName = xmlNodeGetContent(node);
      Log(LOGDEBUG, "Akamai streamName: %s", streamName);

    } else if (xmlStrEqual(node->name, BAD_CAST "authParams")) {
      if (authParams) xmlFree(authParams);
      authParams = xmlNodeGetContent(node);
      Log(LOGDEBUG, "Akamai authparams: %s", authParams);

    } else if (xmlStrEqual(node->name, BAD_CAST "ip")) {
      if (ip) xmlFree(ip);
      ip = xmlNodeGetContent(node);
      Log(LOGDEBUG, "Akamai ip: %s", ip);

    } else if (xmlStrEqual(node->name, BAD_CAST "hostname")) {
      if (hostname) xmlFree(hostname);
      hostname = xmlNodeGetContent(node);
      Log(LOGDEBUG, "Akamai hostname: %s", hostname);

    } else if (xmlStrEqual(node->name, BAD_CAST "appName")) {
      if (appName) xmlFree(appName);
      appName = xmlNodeGetContent(node);
      Log(LOGDEBUG, "Akamai appName: %s", appName);
    }

    node = node->next;
  }

  akamaiStream.Reset();

  if (appName) {
    akamaiStream.app = strdup((char *)appName);
  }

  akamaiStream.protocol = RTMP_PROTOCOL_RTMP;
  akamaiStream.port = 1935;
  akamaiStream.hostname = strdup((char *)hostname);

  if (streamName) {
    int len = xmlStrlen(streamName);
    if (authParams)
      len += xmlStrlen(authParams)+1;
    len++;
    akamaiStream.playpath = (char *)malloc(len*sizeof(char));

    strcpy(akamaiStream.playpath, (char *)streamName);
    if (authParams) {
      strcat(akamaiStream.playpath, "?");
      strcat(akamaiStream.playpath, (char *)authParams);
    }
  }

  if (hostname && appName) {
    akamaiStream.tcUrl = (char *)malloc((7 + xmlStrlen(hostname) + 1 + xmlStrlen(appName) + 1)*sizeof(char));
    strcpy(akamaiStream.tcUrl, "rtmp://");
    strcat(akamaiStream.tcUrl, (char *)hostname);
    strcat(akamaiStream.tcUrl, "/");
    strcat(akamaiStream.tcUrl, (char *)appName);
  }

  if (Link.swfUrl)
    akamaiStream.swfUrl = strdup(Link.swfUrl);
  if (Link.pageUrl)
    akamaiStream.pageUrl = strdup(Link.pageUrl);
  if (Link.SWFHash)
    akamaiStream.SWFHash = strdup(Link.SWFHash);
  akamaiStream.SWFSize = Link.SWFSize;
  if (Link.flashVer)
    akamaiStream.flashVer = strdup(Link.flashVer);
  akamaiStream.seekTime = Link.seekTime;
  akamaiStream.bLiveStream = Link.bLiveStream;
  akamaiStream.timeout = Link.timeout;

  if (ip) xmlFree(ip);
  if (hostname) xmlFree(hostname);
  if (appName) xmlFree(appName);
  if (streamName) xmlFree(streamName);
  if (authParams) xmlFree(authParams);
  xmlFreeDoc(doc);
}
